﻿using System;
using System.Diagnostics;
using System.Globalization;
using System.Runtime.InteropServices;
using System.ComponentModel.Design;
using System.IO;
using Microsoft.Win32;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.OLE.Interop;
using Microsoft.VisualStudio.Shell;
using EnvDTE;
using EnvDTE80;

namespace TafuriSoft.Switcher2012
{
    /// <summary>
    /// This is the class that implements the package exposed by this assembly.
    ///
    /// The minimum requirement for a class to be considered a valid package for Visual Studio
    /// is to implement the IVsPackage interface and register itself with the shell.
    /// This package uses the helper classes defined inside the Managed Package Framework (MPF)
    /// to do it: it derives from the Package class that provides the implementation of the 
    /// IVsPackage interface and uses the registration attributes defined in the framework to 
    /// register itself and its components with the shell.
    /// </summary>
    // This attribute tells the PkgDef creation utility (CreatePkgDef.exe) that this class is
    // a package.
    [PackageRegistration(UseManagedResourcesOnly = true)]
    // This attribute is used to register the information needed to show this package
    // in the Help/About dialog of Visual Studio.
    [InstalledProductRegistration("#110", "#112", "1.0", IconResourceID = 400)]
    // This attribute is needed to let the shell know that this package exposes some menus.
    [ProvideMenuResource("Menus.ctmenu", 1)]
    [Guid(GuidList.guidSwitcher2012PkgString)]
    public sealed class Switcher2012Package : Package
    {
        /// <summary>
        /// Default constructor of the package.
        /// Inside this method you can place any initialization code that does not require 
        /// any Visual Studio service because at this point the package object is created but 
        /// not sited yet inside Visual Studio environment. The place to do all the other 
        /// initialization is the Initialize method.
        /// </summary>
        public Switcher2012Package()
        {
           // Debug.WriteLine(string.Format(CultureInfo.CurrentCulture, "Entering constructor for: {0}", this.ToString()));
        }
        /////////////////////////////////////////////////////////////////////////////
        // Overridden Package Implementation
        #region Package Members

        /// <summary>
        /// Initialization of the package; this method is called right after the package is sited, so this is the place
        /// where you can put all the initialization code that rely on services provided by VisualStudio.
        /// </summary>
        protected override void Initialize()
        {
            Debug.WriteLine(string.Format(CultureInfo.CurrentCulture, "Entering Initialize() of: {0}", this.ToString()));
            base.Initialize();

            // Add our command handlers for menu (commands must exist in the .vsct file)
            OleMenuCommandService mcs = GetService(typeof(IMenuCommandService)) as OleMenuCommandService;
            if (null != mcs)
            {
                // Create the command for the menu item.
                CommandID menuCommandID = new CommandID(GuidList.guidSwitcher2012CmdSet, (int)PkgCmdIDList.cmdidSwitchFiles);
                MenuCommand menuItem = new MenuCommand(MenuItemCallback, menuCommandID);
                mcs.AddCommand(menuItem);
            }
        }

        #endregion

        private void printMessage(string msg, bool focus = false)
        {
            // Retrieve general pane
            IVsOutputWindowPane  generalPane = (IVsOutputWindowPane)GetService(typeof(SVsGeneralOutputWindowPane));

            if (generalPane != null)
            {
                generalPane.OutputString(msg);
                if (focus)
                    generalPane.Activate(); // Brings this pane into view
            }
        }

        private bool findSolutionFile(DTE2 dte, string filename)
        {
            try
            {
                Solution2 solution = (Solution2)dte.Solution;
                ProjectItem p = solution.FindProjectItem(filename);
                if (p == null)
                    return false;
                Window w = p.Open();
                if (w == null)
                    return false;
                w.Activate();
                return true;
            }
            catch (Exception e)
            {
                return false;
            }
        }

        private bool trySwitch(DTE2 dte, string fileName)
        {
            try
            {
                Window opened = dte.OpenFile(fileName, EnvDTE.Constants.vsViewKindCode);
                return opened != null;
            } catch(Exception e) {
                return false;
            }
        }

        private void switchExtensions(string fileName, string[] extensions)
        {
            DTE2 dte = (DTE2)GetService(typeof(EnvDTE.DTE));
            //foreach (string ext in extensions)
            //{
            //    // Test with full path first
            //    string fullPath = Path.ChangeExtension(fileName, ext);
            //    if (trySwitch(dte, fullPath))
            //        return;
            //}
            foreach (string ext in extensions)
            {
                // Otherwise just test with file
                string name = Path.ChangeExtension(Path.GetFileName(fileName), ext);
                if (findSolutionFile(dte, name))
                    return;
            }

            // Failed to find switched file
            printMessage("Failed to switch from file '" + Path.GetFileName(fileName) + "'\n", true);
        }

        private void switchFile(string fileName)
        {
            string fileExt = Path.GetExtension(fileName).ToLowerInvariant();
            switch (fileExt)
            {
                case ".cpp":
                case ".cc":
                case ".c":
                    switchExtensions(fileName, new string[] { ".h", ".hpp" });
                    break;
                case ".h":
                case ".hpp":
                    switchExtensions(fileName, new string[] { ".cpp", ".cc", ".c" });
                    break;
                case ".cu":
                    switchExtensions(fileName, new string[] { ".cuh", ".h" });
                    break;
                case ".cuh":
                    switchExtensions(fileName, new string[] { ".cu"});
                    break;
                default:
                    break;
            }
        }

        /// <summary>
        /// This function is the callback used to execute a command when the a menu item is clicked.
        /// See the Initialize method to see how the menu item is associated to this function using
        /// the OleMenuCommandService service and the MenuCommand class.
        /// </summary>
        private void MenuItemCallback(object sender, EventArgs e)
        {
            DTE2 dte = (DTE2)GetService(typeof(EnvDTE.DTE));
            if (dte.ActiveDocument != null)
            {
                string currentFile = dte.ActiveDocument.FullName;
                if (currentFile.Length > 0)
                    switchFile(currentFile);
            }
        }
    }
}
